/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */
package de.bsvrz.ste.rdstmcvew;
 
import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataNotSubscribedException;
import de.bsvrz.dav.daf.main.OneSubscriptionPerSendData;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SendSubscriptionNotConfirmed;
import de.bsvrz.dav.daf.main.SenderRole;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.dambach.parameter.ParameterEmpfaenger;
import de.bsvrz.sys.funclib.dambach.parameter.ParameterSender;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Die Klasse zum Versenden von Daten unter dem Aspekt "Senden" der
 * Attributgruppe "RDSMeldung". 
 * Sie implementiert das ClientSenderInterface welches die Methoden
 * zum Versenden der Daten bereitstellt.
 * 
 * @author Dambach Werke GmbH
 * @author Stefan Sans
 * @version $Revision: 1.2 $ / $Date: 2008/09/30 15:20:19 $ / ($Author: Hilpp $)
 * 
 * Geprüft Hilpp 30.9.2008
 */
public class RDSVewSender implements ClientSenderInterface
{
  /**
   * Debug-Logger für Logging-Ausgaben
   */
  private final static Debug _debug = Debug.getLogger();

  /**
   * Ist auf true gesetzt, solange auf Sendesteuerung gewartet wird
   */
  private boolean _warten = false;

  /**
   * Rolle, mit der als Sender angemeldet wird
   */
  public static final SenderRole ROLLE = SenderRole.source();

  /**
   * Zeitin ms, die max. auf Sendesteuerung gewartet wird
   */
  public static final long SENDE_TIMEOUT = 5000;

  /**
   * Ist diese Instanz zum DAV-Senden angemeldet?
   */
  private boolean _angemeldet = false;

  /**
   * @return ist diese Instanz zum DAV-Senden angemeldet?
   */
  public boolean isAngemeldet()
  {
    return _angemeldet;
  }

  /**
   * Datenbeschreibung "gesendet"
   */
  protected final DataDescription _datenBeschreibung;

  /**
   * Übergeordnet Instanz vom Typ RDSVewMeldung
   */
  protected final RDSVewMeldung meldung;

  /**
   * @return die Datenverteiler-Verbindung
   */
  public ClientDavInterface getConnection()
  {
    return meldung.getConnection();
  }

  /**
   * @return das Dav-Systemobject, für das das Senden von Daten erfolgt
   */
  public SystemObject getDavObjekt()
  {
    return meldung.getDavObjekt();
  }

  /**
   * Konstruktor der Klasse
   *
   * @param meldung Objekt vom Typ RDSVewMeldung,
   * für das das Senden von Daten erfolgt
   */
  public RDSVewSender(RDSVewMeldung meldung)
  {
    this.meldung = meldung;

    AttributeGroup atg = getConnection().getDataModel().getAttributeGroup("atg.rdsMeldung");
    Aspect asp = getConnection().getDataModel().getAspect("asp.rdsSenden");
    _datenBeschreibung = new DataDescription(atg, asp);

    /*
     * Restauriere unter dem Aspekt RDSSenden gesendete Daten aus den Parametern -
     * da diese Daten als Quelle versendet wurden
     */
    _datenSenden = new ParameterEmpfaenger(getConnection(), getDavObjekt(),
        "atg.rdsMeldungSenden", "atg.rdsMeldung").getData();

    try
    {
      getConnection().subscribeSender(this, getDavObjekt(), _datenBeschreibung, ROLLE);
      _debug.fine(_datenBeschreibung + " zum Senden angemeldet für " + getDavObjekt());
    }
    catch (OneSubscriptionPerSendData e)
    {
      /* Es liegt bereits eine Sendeanmeldung von anderer Stelle vor
       * => hier nicht anelden
       */
      _debug.warning(meldung.getDavObjekt().getPid() +
          " ist bereits an anderer Stelle zum Senden angemeldet", e);
      return;
    }
    _angemeldet = true;
  }

  /**
   * Gibt die mit dieser Instanz verbundenen Ressourcen wieder frei
   */
  public void dispose()
  {
    stopSenden();
    if (_angemeldet)
    {
      getConnection().unsubscribeSender(this, getDavObjekt(), _datenBeschreibung);
      _debug.fine("Senden abgemeldet von " + _datenBeschreibung + " für " + getDavObjekt());
      _angemeldet = false;
    }
  }

  /**
   * Unmittelbares Senden von Daten.
   *
   * @param resultData die zu sendenden Daten
   * @return true, wenn das Senden erfolgreich war, false sonst
   */
  protected boolean send(ResultData resultData)
  {
    try
    {
      getConnection().sendData(resultData);
    }
    catch(SendSubscriptionNotConfirmed e)
    {
      if (SENDE_TIMEOUT > 0)
      {
        _debug.fine("Warte max. " + SENDE_TIMEOUT + "ms auf Sendesteuerung für " + getDavObjekt().getPid());
        synchronized(this)
        {
          _warten = true;
          try
          {
            wait(SENDE_TIMEOUT);
          }
          catch(InterruptedException ex)
          {
            // tue nichts
          }
          _warten = false;
        }
        try
        {
          getConnection().sendData(resultData);
        }
        catch (SendSubscriptionNotConfirmed ex)
        {
          _debug.warning("Keine Sendesteuerung für " + getDavObjekt().getPid(), ex);
          return false;
        }
        return true;
      }
      else
      {
        _debug.warning("Keine Sendesteuerung für " + getDavObjekt().getPid(), e);
        return false;
      }
    }
    catch (DataNotSubscribedException e)
    {
      _debug.error("Keine Sendeanmeldung bzw. Sendeanmeldung ungültig für " + getDavObjekt().getPid(), e);
      return false;
    }
    return true;
  }

  /*
   * (Kein Javadoc)
   * @see de.bsvrz.dav.daf.main.ClientSenderInterface#dataRequest(de.bsvrz.dav.daf.main.config.SystemObject, de.bsvrz.dav.daf.main.DataDescription, byte)
   */
  public void dataRequest(SystemObject object, DataDescription dataDescription, byte state)
  {
    _debug.fine("dataRequest() für " + getDavObjekt().getPid() + " - Sender state: " + state);
    if (_warten && state == ClientSenderInterface.START_SENDING)
    {
      synchronized(this)
      {
        notify();
      }
    }
  }

  /*
   * (Kein Javadoc)
   * @see de.bsvrz.dav.daf.main.ClientSenderInterface#isRequestSupported(de.bsvrz.dav.daf.main.config.SystemObject, de.bsvrz.dav.daf.main.DataDescription)
   */
  public boolean isRequestSupported(SystemObject so, DataDescription dd)
  {
    return true;
  }

  /**
   * Aktuell unter dem Aspekt senden gesendete Daten
   */
  private ResultData _datenSenden = null;

  /**
   * @return die unter dem Aspekt senden aktuell gesendeten Daten
   */
  public ResultData getDatenSenden()
  {
    return _datenSenden;
  }

  /**
   * setzt die unter dem Aspekt senden aktuell gesendeten Daten
   *
   * @param datenSenden die aktuell gesendeten Daten
   */
  void setDatenSenden(ResultData datenSenden)
  {
    _datenSenden = datenSenden;
  }

  /**
   * Aktuell aktiver SendeThread
   */
  private SendeThread _sendeThread = null;

  /**
   * @return den aktuell aktiven SendeThread
   */
  SendeThread getSendeThread()
  {
    return _sendeThread;
  }

  /**
   * Setzt den aktuell aktiven SendeThread
   *
   * @param sendeThread der zu setztende SendeThread
   */
  void setSendeThread(SendeThread sendeThread)
  {
    _sendeThread = sendeThread;
  }

  /**
   * Starte neuen SendeTread - ein evtl. laufender Sendethread wird durch
   * den übergebenen ersetzt und ggf. abgebrochen
   *
   * @param data die zu sendenden Daten - der Sendeauftrag verwaltet eine Kopie
   * dieser Daten
   * @param verzoegerung die Sendeverzögerung in s
   * @param zeitStempel ZeitStempel
   * @param bedingterAbbruch wenn <code>true</code> wird ein laufender SendeThread
   * abgebrochen, wenn er mit Verzögerung gestartet wurde
   * @param empfangsAspekt EmpfangsAspekt
   */
  public void starteSenden(Data data, long zeitStempel, long verzoegerung,
      boolean bedingterAbbruch, Aspect empfangsAspekt)
  {
    if (bedingterAbbruch)
    {
      if (_sendeThread != null &&
          _sendeThread.empfangsAspekt.equals(meldung.datenBeschreibungGeneriert.getAspect()))
      {
        stopSenden();
        _debug.fine("Laufenden Sende-Auftrag abgebrochen für " + getDavObjekt().getPid());
      }
    }
    synchronized(this)
    {
      _sendeThread = new SendeThread(data, zeitStempel, verzoegerung, empfangsAspekt);
      _sendeThread.start();
    }
  }

  /**
   * Abbrechen eines evtl. gestarteten und laufenden SendeThreads
   */
  public void stopSenden()
  {
    synchronized(this)
    {
      if (_sendeThread != null)
      {
        _sendeThread.abbrechen();
      }
    }
  }

  /**
   * Senden eines Datensatzes mit optionaler Verzögerung in einem
   * eigenen Thread
   *
   * @author Stefan Sans
   */
  class SendeThread extends Thread
  {
    /**
     * Sende-Verzögerungszeit in ms
     */
    final long verzoegerung;

    /**
     * Datensatz, der gesendet werden soll
     */
    final Data data;

    /**
     * Zeitstempel der Daten
     */
    final long zeitStempel;

    /**
     * Aspekt, unter dem die zu sendenden Daten ursprünglich empfangen wurden
     */
    final Aspect empfangsAspekt;

    /**
     * Flag, das angibt, ob der Thread abgebrochen wurde
     */
    private boolean _abgebrochen = false;

    /**
     * @return wurde der SendeThread abgebrochen?
     */
    boolean isAbgebrochen()
    {
      return _abgebrochen;
    }

    /**
     * Konstruiert einen neuen SendeThread zum Senden eines Datensatzes
     * mit optionaler Sendeverzögerung
     *
     * @param data der Datensatz, der gesendet werden soll
     * @param verzoegerung Sendeverzögerung in ms
     * 
     */
    SendeThread(Data data, long zeitStempel, long verzoegerung, Aspect empfangsAspekt)
    {
      super();
      this.data = data;
      this.zeitStempel = zeitStempel;
      this.verzoegerung = verzoegerung;
      this.empfangsAspekt = empfangsAspekt;
    }
    
    /**
     * Abbrechen des laufenden Threads
     */
    void abbrechen()
    {
      synchronized(this)
      {
        interrupt();
        _abgebrochen = true;
      }  
    }
    
    @Override
    public void run()
    {
      if (!isAbgebrochen() && verzoegerung > 0)
      {
        getConnection().sleep(verzoegerung);
      }
      synchronized(RDSVewSender.this)
      {
        try
        {
          if (!isAbgebrochen())
          {
            String text = "Daten mit Zeitstempel " + zeitStempel +
            " empfangen unter dem Aspekt " + empfangsAspekt.getName();
            
            if (empfangsAspekt.equals(meldung.datenBeschreibungGeneriert.getAspect()) && 
                meldung.getQuelle(data) == RDSVewMeldung.Quelle.Automatisch &&
                meldung.getZustand(data) == RDSVewMeldung.Zustand.NichtQuittiert &&
                meldung.getRDSMeldungsAktion().equals("0"))
            {
              _debug.fine(text + " werden NICHT unter dem Aspekt " + 
                _datenBeschreibung.getAspect().getName() + 
                " gesendet, da RDSMeldungsAktion den Wert \"" + 
                meldung.getRDSMeldungsAktion() + "\" hat");
              return;
            }              

            ResultData resultData = new ResultData(getDavObjekt(), _datenBeschreibung, 
                zeitStempel, data);
            
            if (send(resultData))
            {
              _debug.fine(text + " wurden erfolgreich unter dem Aspekt " + 
                  _datenBeschreibung.getAspect().getName() + " gesendet"); 
                   
              setDatenSenden(resultData);
              
              /*
               * Persistiere unter dem Aspekt RDSSenden gesendete Daten als Parameter -
               * da diese Daten als Quelle versendet wurden
               */
              new ParameterSender(getConnection(), getDavObjekt(), 
                  resultData, "atg.rdsMeldungSenden").start();
            }  
          }
        }
        finally
        {
          if (getSendeThread() == this)
          {
            setSendeThread(null);
          }
        }  
      }
    }
    
  }
 
}
